#include <stdio.h>
#include <string.h>
#include <ctype.h> 
#include <assert.h>

#define MAX_STACK 100
#include "global-stack.h"
#include "tokens.h"

/* --------- ERROR MESSAGES */
#define printerr(format, ...) printf("error: " format "\n", ##__VA_ARGS__)
#define UNKNOWN_TOKEN_ERR "unknown token %s"
#define VAR_NOT_SET_ERR "variable %c has no value"


/* --------- PROMPT FORMATS */
#define PROMPT_F     ">>> "
#define POP_PRINT_F  "%.8g\n"
#define PEEK_PRINT_F "-> %.8g\n"
#define INITIAL_PROMPT_F \
    "A simple polish notation calculator.\n" \
    "Type 'help' for a list of available commands\n" \
    PROMPT_F

/* -------- STACK PRINTING FUNCTIONS */
double lst_print;
/* print_peek: print top element, leave stack unchanged */
#define print_peek() printf(PEEK_PRINT_F, lst_print = peek())
/* print_pop: pop top element and print it */
#define print_pop() printf(POP_PRINT_F, lst_print = pop())


/* ------- GLOBAL VARIABLE NAMESPACE */
#define NS_SIZE ('z' - 'a' + 1)
    
char var_flags[NS_SIZE];
double var_vals[NS_SIZE];

/* is_var: return 0 if c is not a variable name */
#define is_var(c) isalpha(c)

/* set_var: pop val from stack, set variable c to val */
#define set_var(c) \
    do { \
        int i = tolower(c) - 'a'; \
        var_flags[i] = 1; \
        var_vals[i] = pop(); \
    } while (0)

/* get_var: push value of c to stack */
#define get_var(c) \
    do { \
        int i = tolower(c) - 'a'; \
        if (!var_flags[i]) \
            printerr(VAR_NOT_SET_ERR, i+'a'); \
        else \
            push(var_vals[i]); \
    } while (0)

/* print_vars: print all set variables and their values */
void print_vars(void)
{
    int i;
    for (i = 0; i < NS_SIZE; i++)
        if (var_flags[i])
            printf("%c = %.8g\n", i+'a', var_vals[i]);
}

/* variable manipulation tokens */
#define VAR_TOKENS \
    X('=', "pop value top from stack and set %s to value", set_var) \
    X('$', "push value of %s to top of stack", get_var)


/* --------- MAIN PROGRAM */ 
#define MAXTOKEN 100
int gettoken(char *token, int max_t);
char * parsef(char *s, double *x);
void help(void);

/* reverse Polish calculator */
int main()
{
    char token[MAXTOKEN + 1];
    double x;
    int t_len;
    
    printf(INITIAL_PROMPT_F);
    while ((t_len = gettoken(token, MAXTOKEN))) {
        if (t_len < 0) { /* newline token */
            if (length())
                print_pop();
            printf(PROMPT_F);
        } else
    #define X(F, DESC, APPLY, ...) \
        if (t_len == sizeof(#F) / sizeof(char) - 1 && !strcmp(#F, token)) \
            APPLY(F, ##__VA_ARGS__); \
        else
        TOKENS
    #undef X
    #define X(C, DESC, F) \
        if (t_len == 2 && token[0] == C && is_var(token[1])) \
            F(token[1]); \
        else
        VAR_TOKENS
    #undef X
        if (*parsef(token, &x) == '\0')
                    push(x);
        else
            printerr(UNKNOWN_TOKEN_ERR, token);
    }

    putchar('\n');
    while (length())
        print_pop();

    return 0;
}


/* ---------- HELP COMMAND */
#define MAX_CMD_W 10 /* token padding */
#define VAR_STR "X"  /* string to represent variable */

/* we verify that no token is to large to be pretty printed */
#define X(TOKEN, ...) \
    static_assert(sizeof(#TOKEN) / sizeof(char) <= MAX_CMD_W, \
        "size of " #TOKEN " is to large for pretty printing " \
        "in help command");
    TOKENS
#undef X

/* help: prints all available calculator functions */
void help(void)
{
#define X(C, DESC, ...) \
    printf("%c%-*s" DESC "\n", C, MAX_CMD_W-1, VAR_STR, VAR_STR);
    VAR_TOKENS
#undef X
#define X(F, DESC, ...) printf("%-*s" DESC "\n", MAX_CMD_W, #F);
    TOKENS
#undef X
}


/* -------- PARSING FUNCTIONS */
#define BLANK ' '

/* gettoken: get next space delimited token from input
 * return token length, -1 on newline and 0 on EOF */
int gettoken(char *token, int t_max) 
{
    static int c = BLANK;
    
    while (isspace(c) && c != '\n') /* remove leading whitespace */
        c = getchar();

    if (c == '\n') { /* newline is a token */
        c = BLANK;
        return -1;
    }

    /* fetch token */
    int i;
    for (i = 0; i < t_max && c != EOF && !isspace(c); c = getchar(), i++) 
        token[i] = c;

    token[i] = '\0';
    return i;
}

#define DEC_SEP X('.') X(',')
#define EXP_SEP X('e') X('E')

/* macros for parsef */
#define SIGN(s) \
    ((*s == '-') ? (s++, -1) : ((*s == '+') ? (s++, 1) : 1))
#define ATOI(s, x, ...) \
    for (; isdigit(*s); s++, ##__VA_ARGS__) \
        (x) = 10 * (x) + (*s - '0')

/* parsef: convert string s to double, write it to *x
 * return pointer of last read char in s  */
char * parsef(char s[], double *x)
{
    *x = 0;
    /* integer part */
    int sign = SIGN(s);    
    ATOI(s, *x);

    int exp = 0; 
    /* fractional part */
#define X(sep) *s == sep || 
    if (DEC_SEP 0) {
        s++;
        ATOI(s, *x, exp--);
        *x *= pow(10, exp);
    }
    *x *= sign;
#undef X
    /* exponent part */
#define X(sep) *s == sep || 
    if (EXP_SEP 0) {
        s++;
        sign = SIGN(s);
        exp = 0;
        ATOI(s, exp);
        *x *= pow(10, sign * exp);
    }
#undef X
    return s;
}
